/** \file
	cntconv
	2004, Andreas Romeyke
	mailto://art1@it-netservice.de

	based on cntread
	reads a compressed EEG-stream,
	decompressed it, and extracts m-Channels as a binary stream to stdout
	normalized by dividing with 2^16, because EEG-devices have a resolution of
	22 bits in a range of plusminus 2 Volt. All values greater 1 will be saturated to 1.
*/



#include "local.h"
#include "cnt.h"

static char cntconv_version[]="$Id: cntconv.c,v 1.15 2004/08/05 09:56:00 romeyke Exp $";

/** this Macro is needed for hash_ten_twenty() */
#define ERRHASHTENTWENTYMACRO { \
		/* fprintf(stderr, "left-string \"%s\" in chn-file not of following string-constants:", left); \
		fprintf(stderr, "eogh, eogv, pg1, pg2, fp1, fp2, f7, f8, f3, fz, f4, t3, c3, cz, c4, t4, a1, a2, p3, pz, p4, t5, t6, o1, o2, empty\n"); */ \
		return (-1); \
}

/** returns the corresponding enumeration of type e_channelmap of an given string, if string does not exist a value of -1 will be returned. It is not really a hashing function but a binary search of a set of strings corresponding enumeration described above. */
static int hash_ten_twenty(const  char * left) {
	enum e_channelmap i;
	i=empty; /* enum */
	if (strlen(left)>5) {
		ERRHASHTENTWENTYMACRO;
	}
	if (strlen(left)==2) {
		switch (left[0]) {
			case 'a':
				switch (left[1]) {
					case '1': /* a1 */
						i=a1; break;
					case '2': /* a2 */
						i=a2; break;
					default:
						ERRHASHTENTWENTYMACRO; break;
				}
				break;
			case 'c':
				switch (left[1]) {
					case '3':
						i=c3; break;
					case 'z':
						i=cz; break;
					case '4':
						i=c4; break;
					default:
						ERRHASHTENTWENTYMACRO; break;
				}
				break;
			case 'f':
				switch (left[1]) {
					case '7': /* f7 */
						i=f7; break;
					case '8': /* f8 */
						i=f8; break;
					case '3': /* f3 */
						i=f3; break;
					case 'z': /* fz */
						i=fz; break;
					case '4': /* f4 */
						i=f4; break;
					default: ERRHASHTENTWENTYMACRO; break;
				}
				break;
			case 'o':
				switch (left[1]) {
					case '1':
						i=o1; break;
					case '2':
						i=o2; break;
					default:
						ERRHASHTENTWENTYMACRO; break;
				}
				break;
			case 'p':
				switch (left[1]) {
					case '3':
						i=p3; break;
					case 'z':
						i=pz; break;
					case '4':
						i=p4; break;
					default:
						ERRHASHTENTWENTYMACRO;; break;
				}
				break;
			case 't':
				switch (left[1]) {
					case '3':
						i=t3; break;
					case '4':
						i=t4; break;
					case '5':
						i=t5; break;
					case '6':
						i=t6; break;
					default:
						ERRHASHTENTWENTYMACRO; break;
				}
				break;
			default:
				ERRHASHTENTWENTYMACRO; break;
		}
	} else if (strlen(left)==3) {
		if (0==strcmp("fp1", left)) i=fp1;
		else if (0==strcmp("fp2", left)) i=fp2;
		else { ERRHASHTENTWENTYMACRO; }
	} else if (strlen(left)==4) {
		if (0==strcmp("eogh", left)) i=eogh;
		else if (0==strcmp("eogv", left)) i=eogv;
		else { ERRHASHTENTWENTYMACRO; }
	} else if (strlen(left)==5) {
		if (0==strcmp("empty", left)) i=empty;
		else { ERRHASHTENTWENTYMACRO; }
	} else { ERRHASHTENTWENTYMACRO; }
	return (int) i;
}

/** parse_chn parses the chn-Files/channel-mapping. It needs a filename with channelmapping. The channels in a given cnt-file will be mapped via *.chn-file into the International 10-20-system. The left-side were the 10-20-labels, the right-side the labels of the cnt-file.  */
static int parse_chn(eeg_t * cnt, char * filename){
	FILE * CHIN;
	int tmpleft;
	short int index;
	char tmpstring[20];
	int tmpcheck;
	char * tmpleftstring;
	char * ret;
	char * p;
	/* open the CHIN */
	if (NULL==filename) {
		fprintf(stderr, "there was no filename specified for CHN-data, %s\n", strerror(errno));
		exit(ERRSTREAM);
	}
	CHIN = fopen(filename, "r");
	if (CHIN == NULL) {
		fprintf(stderr, "cannot open %s, %s\n", filename, strerror(errno));
		exit(ERRSTREAM);
	}
	tmpcheck=0;
	while (
		(tmpcheck!=0x7fffff)
		){
		if (0==feof(CHIN)) {
			ret=fgets(tmpstring, 19, CHIN);
#ifdef DEBUG
			fprintf(stderr, "debug: %u %s\n", i, ret);
#endif
			if (0!=feof(CHIN)) {
				fprintf(stderr, "unexpected EOF in file %s, needs more entries (from %X only %X bits with enumerated labels set), %s\n", filename, 0x1ffffff, tmpcheck, strerror(errno));
				exit(ERRSTREAM);
			}
			if (ret==NULL) {
				fprintf(stderr, "There was an error reading file %s, %s\n", filename, strerror(errno));
				exit(ERRSTREAM);
			}
			if (ret[0]=='#') continue;
			p=strtok(ret, "=");
			if (p==NULL) {
				fprintf(stderr, "error in parsing file %s, wrong right value, %s\n", filename, strerror(errno));
				exit(ERRSTREAM);
			}
			tmpleft=hash_ten_twenty(p);
			tmpleftstring=p;
			if (tmpleft== -1) {
				fprintf(stderr, "wrong left-values in file %s\n", filename);
				exit(ERRSTREAM);
			}
			p=strtok(NULL, "\n");
			if (p==NULL) {
				fprintf(stderr, "error in parsing file %s, %s\n", filename, strerror(errno));
				exit(ERRSTREAM);
			}
			index=get_chan_index(cnt, p);
			if (index == -1) {
				fprintf(stderr, "error in parsing file %s, wrong right-label, label %s must exist in cnt-file, %s\n", filename, p, strerror(errno));
				exit(ERRSTREAM);
			}
			tmpcheck=tmpcheck|(1<<(tmpleft-1)) ;
#ifdef DEBUG
			fprintf(stderr, "Leftlabel: %s, Rightlabel: %s, tmpleft: %i tmpcheck:%x\n", tmpleftstring, p, tmpleft,tmpcheck);
#endif
			set_chan_lab(cnt,index, tmpleftstring);
		} else {
			fprintf(stderr, "unexpected EOF in file %s, needs more entries (from %X only %X bits with enumerated labels set), %s\n", filename, 0x7fffff, tmpcheck, strerror(errno));
			exit(ERRSTREAM);
		}
	}
	(void) fclose(CHIN);
	return(0);
}

/**
Decompresses a cnt-file, maps the channel via a chn-File, does normalize the EEG-data and writes a binary (with option "-b") oder human-readable (with option "-h") stream to stdout.
If the option "-d" is used there will be channel-labels of existing cnt-File printed to STDOUT.

The filename is the filename without suffix, the suffixes ".cnt" and ".chn" will be added automatically.

usage: cntconv -[hbd] <cnt-File>
*/
int main(int argc, char *argv[]) {
	unsigned short int channelcount;
	unsigned int samplecount;
	float * pcbuf;
	unsigned short int * prbuf;
	unsigned short int outputformat;
	pcbuf=NULL;
	prbuf=NULL;
	outputformat=0;

	if ((argc != 3)||
		((strcmp(argv[1], "-h")!=0) &&
		 (strcmp(argv[1], "-b")!=0) &&
		 (strcmp(argv[1], "-d")!=0)
		)
	) {
		fprintf(stderr, "usage: %s -[hbd] <cnt-File>\n", argv[0]);
		fprintf(stderr, "\t -b means binary output\n\t -h means human readable output\n");
		fprintf(stderr, "\t -d dumps the channel-labels of actual <cnt-File>\n");
		fprintf(stderr, "\twhere <cnt-File> contains the compressed EEG-data\n");
		fprintf(stderr, "\tthe cnt-files must also have at least %i channels\n", MAX_MAPPED_CHANNELS);
		exit(ERRARG);
	}
	if (fflush(stdout)!=0) {
		fprintf(stderr, "An error occured by flushing streams: %s", strerror(errno));
		exit(ERRSTREAM);
	}
	if (strcmp(argv[1], "-d")==0) {
		(void) print_chanlabels(argv[2]);
	} else {
		/* first parse_cnt, second parse_rej, because rej depends on samplecount */
		(void) parse_cnt(argv[2], &channelcount, &samplecount, &pcbuf); /* parse EEG-data, data in pcbuf */
		assert(pcbuf!=NULL);
		if (strcmp(argv[1], "-b")==0) outputformat=0; /* binary */
		else outputformat=1; /* human */
		(void) write_header(&channelcount, &samplecount, outputformat);
		(void) write_data(&channelcount, &samplecount, &pcbuf, outputformat);
	}
	return(ERROK);
}

/** writes the decompressed data from buffer to a file. ppcbuf is a pointer to the startpointer of this buffer. */
int write_data(unsigned short int * chanc, unsigned int * samplec, float ** ppcbuf, unsigned short int outputformat) {
	int channelcount;
	int samplecount;
	int channel;
	int sample;
	float * pcbuf;
	int ret;
	assert(chanc!=NULL);
	assert(samplec!=NULL);
	assert(ppcbuf!=NULL);
	pcbuf= *ppcbuf;
	assert(pcbuf!=NULL); /* because it should be alloc */
	channelcount= (int) *chanc;
	assert(channelcount>0);
	fprintf(stderr, "channelcount is: %i\n", channelcount);
	assert(channelcount==MAX_MAPPED_CHANNELS);
	samplecount= (int) *samplec;
	assert(samplecount>0);
	ret=0;
	switch (outputformat) {
		case 0: /* binary */
		/* DONE: fixed BUG, in binary there was "bad address"-error in prbuf */
		/* TODO: for higher portability we should not write the structure at once, but its elements, check also parseeeg.c */
			ret=(int)fwrite(pcbuf, sizeof(float), (size_t) (channelcount*samplecount), stdout);
			if (ret != (int) (channelcount*samplecount)) {
				fprintf(stderr, "could not write pcbuf to stdout-stream, only %i of %i written: %s\n", ret, (channelcount*samplecount),strerror(errno));
				exit(ERRSTREAM);
			}
			break;
		case 1: /* human */
			printf("# sample, artifact, channel1, channel2,...,channelN\n");
			for (sample=0; sample<samplecount; sample++) {
				printf("%i", sample+1);
				for (channel=0; channel<channelcount; channel++) {
					printf(" %f",pcbuf[channel*channelcount+sample]);
				}
				printf("\n");
			}
			break;
		default: /* should never passed */
			fprintf(stderr, "error in code at line %i\n", __LINE__);
			exit(ERRPRGCODE);
	}
	return(0);
}

/** write out header, needs channel-count and sample-count */
int write_header(unsigned short int *chanc, unsigned int *samplec, unsigned short int outputformat) {
	unsigned short int channelcount;
	unsigned int samplecount;
	struct s_cntconvhead header;
	int ret;
	channelcount= *chanc;
	samplecount= *samplec;
	/* write header */
	header.channelcount=channelcount;
	header.samplecount=samplecount;
	header.norm=simple;
	switch (outputformat) {
		case 0: /* binary */
			printf("#EEGbu");
			ret=(int)fwrite(&header, sizeof(header), (size_t) (1), stdout);
			if (ret != (int) (1)) {
				fprintf(stderr, "could not write pcbuf to stdout-stream, only %i of %i written: %s\n", ret, (1),strerror(errno));
				exit(ERRSTREAM);
			}
			break;
		case 1: /* human */
			printf("#EEG Data stream uncompressed\n");
			printf("#id=%i\n", 1); /* means human readable */
			printf("#(in a row all channels, in col all samples)\n");
			printf("#channels=%u\n",header.channelcount);
			printf("#samples=%u\n", header.samplecount);
			printf("#norm function=%i\n",header.norm);
			printf("#normvalue=%u\n", NORMVALUE);
			printf("#format=littleEndian, float, %u Bytes\n", (unsigned int) sizeof(float));
			printf("#data=\n");
			break;
		default: /* should never passed */
			fprintf(stderr, "error in code at line %i\n", __LINE__);
			exit(ERRPRGCODE);
	}
	return(0);
}


/** prints the labelnames in cnt-File to stdout to support easier creating of channelmapping-file */
int print_chanlabels(char * filename) {
	FILE * CIN;
	eeg_t *src;
	char * cntfilename;
	int status;
	unsigned short int channelcount;
	cntfilename=malloc((strlen(filename)+5)*sizeof(char));
	if (NULL==cntfilename) {
		fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	strcpy(cntfilename, filename);
	strcat(cntfilename, ".cnt");
	CIN = fopen(cntfilename, "rb");
	if (CIN == NULL) {
		fprintf(stderr, "cannot open %s, %s\n", cntfilename, strerror(errno));
		exit(ERRSTREAM);
	}
	src = cnt_init_file(cntfilename, CIN, &status);
	if (src==NULL) {
		fprintf(stderr, "error reading %s, status is %i\n", cntfilename, status);
		exit(ERRSTREAM);
	}
	channelcount = (unsigned short int) get_cnt_chanc(src);      /* number of channels in cnt */
	if (channelcount < REQUIREDCHANNELS) {
		fprintf(stderr, "File %s has only %u channels, but at least %u required\n", filename, channelcount, REQUIREDCHANNELS);
		exit(ERRCHANCOUNT);
	}
	if (cntclose(src)!=0) { /* release src */
		fprintf(stderr, "could not free structures of file %s, error of cntopenlib\n", filename);
		exit(ERRSTREAM);
	};
	if (fclose(CIN)!=0) {
		fprintf(stderr, "could not close file %s, error is %s\n", filename, strerror(errno));
		exit(ERRSTREAM);
	}
	free(cntfilename);
	return(ERROK);
}

/** decompresses EEG-data from a cnt-file and returnes the channelcount, samplecount and a new buffer of floats with EEG-data, ppcbuf is a pointer to the startpointer of this buffer */
int parse_cnt(char * filename, unsigned short int *chanc, unsigned int *samplec, float ** ppcbuf) {
	FILE * CIN;
	eeg_t *src;
	int status;
	sraw_t *buf;
	float * pcbuf;
	unsigned short int channelcount, chan;
	int chan_mapped;
	unsigned int samplecount, sample;
	float normv;
	char * cntfilename;
	char * chnfilename;

	pcbuf=NULL;
	assert(ppcbuf!=NULL); /* should not point to memory */
	assert(chanc!=NULL); /* we return values, a real address is needed */
	assert(samplec!=NULL); /* we return values, a real address is needed */
	pcbuf= *ppcbuf;
	samplecount= *samplec;
	channelcount= *chanc;
	status=0;
	buf=NULL;
	src=NULL;
	cntfilename=malloc((strlen(filename)+5)*sizeof(char));
	if (NULL==cntfilename) {
		fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	strcpy(cntfilename, filename);
	strcat(cntfilename, ".cnt");
	chnfilename=malloc((strlen(filename)+5)*sizeof(char));
	if (NULL==chnfilename) {
		fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	strcpy(chnfilename, filename);
	strcat(chnfilename, ".chn");

	CIN = fopen(cntfilename, "rb");
	if (CIN == NULL) {
		fprintf(stderr, "cannot open %s, %s\n", cntfilename, strerror(errno));
		exit(ERRSTREAM);
	}
	src = cnt_init_file(cntfilename, CIN, &status);
	if (src==NULL) {
		fprintf(stderr, "error reading %s, status is %i\n", cntfilename, status);
		exit(ERRSTREAM);
	}
	channelcount = (unsigned short int) get_cnt_chanc(src);      /* number of channels in cnt */
	if (channelcount < REQUIREDCHANNELS) {
		fprintf(stderr, "File %s has only %u channels, but at least %u required\n", filename, channelcount, REQUIREDCHANNELS);
		exit(ERRCHANCOUNT);
	}

	(void) parse_chn(src, chnfilename); /* parse channel-file to get channel mapping to 10-20-system */
	/* obtain some attributes */
	samplecount = (unsigned int) get_cnt_samplec(src);  /* number of samples in cnt */
	*samplec=samplecount;
	buf = (sraw_t *) malloc((channelcount + 2) * sizeof(sraw_t));
	if (buf==NULL) {
		fprintf(stderr, "Could not allocate memory for buffer\n");
		exit(ERRMEM);
	}
	/* TODO: in case of avoiding mem-leaks, there should be added a function to release this buffer at the end of main() */
	pcbuf = (float *) malloc(MAX_MAPPED_CHANNELS*samplecount*sizeof(float));
	if (pcbuf==NULL) {
		fprintf(stderr, "Could not allocate memory for buffer\n");
		exit(ERRMEM);
	}

	for (sample = 0; sample < samplecount; sample++) {
		if (cntread(src, buf, 1)!=0) {
			fprintf(stderr, "error reading %s\n", filename);
			exit(ERRSTREAM);
		} else {
			/*
				buf contains 1 sample per channel, real-world scaled
			*/
			for (chan=0; chan < channelcount; chan++) { /* print */
				chan_mapped=hash_ten_twenty(get_chan_lab(src, chan));
				if (chan_mapped != -1) { /* wenn ok gemappter Channel, setze chan_mapped, sonst berspringen */
					normv=(float) (buf[chan]*get_chan_scale(src, (short int) chan)/(NORMVALUE*MAXTAPS*REQUIREDCHANNELS)); /* real world scaled, OK */
					if (normv >(float) (20)) {
						normv=(float) (20);
						fprintf(stderr, "normv: %f\n", normv);
					}
					if (normv < (float) (-20)) {
						normv= (float) -(20);
						fprintf(stderr, "normv: %f\n", normv);
					}
					pcbuf[sample*MAX_MAPPED_CHANNELS+(chan_mapped-1)]=normv;
				}
			}
			/* printf("\n"); */
		}
	}
	*ppcbuf = pcbuf; /* return data */
	if (cntclose(src)!=0) { /* release src */
		fprintf(stderr, "could not free structures of file %s, error of cntopenlib\n", filename);
		exit(ERRSTREAM);
	};
	free(buf);
	buf=NULL;
	if (fclose(CIN)!=0) {
		fprintf(stderr, "could not close file %s, error is %s\n", filename, strerror(errno));
		exit(ERRSTREAM);
	}
	free(cntfilename);
	free(chnfilename);
	*chanc=MAX_MAPPED_CHANNELS;
	return(0);
}

