/** \file
	ear - EEG artifacts recognizer
	(c) 2004 Andreas Romeyke

	a neural network to reject artifacts with EEG-data
	Version $Id: ear.c,v 1.23 2004/08/05 09:59:04 romeyke Exp $

	it uses
	Fast artificial neural network library (http://fann.sourceforge.net)...

	- load "network-data", read input-data, calc output-data
*/

#include "local_ann.h"

static char ear_version[]="$Id: ear.c,v 1.23 2004/08/05 09:59:04 romeyke Exp $";

/** smoothness of artefakt-detection, value+1, because one indexed value will be substracted from sum */
#define AVGWINDOW (16+1)
/** to correct the offset of detected artifacts in time-scale */
#define OFFSET (128-30-13)
/** defines the threshold if there was an artifact or not */
#define AVGTHRESHOLD (0.5f)

int print_learncurve(int step, float value) {
	FILE * f;
	f=fopen("learncurve.dat", "a");
	if (f==NULL) {
		fprintf(stderr, "File %s could not be open, %s (line %i)\n", "learncurve.dat", strerror(errno), __LINE__);
		exit(ERRSTREAM);
	}
	fprintf(f, "%i %0.10f\n", step, value);
	fclose(f);
}

/** feed it with eeg-data and a trained ANN it outputs a rej-file with artifact-stamps */
int main(int argc, char * argv[]) {
	struct fann * ann;
	unsigned int l;
	unsigned long int tmp;
	unsigned int sample, taps;
	unsigned short int was_an_artifact;
	unsigned short int artifact_count;
	unsigned short int nonartifact_count;
	float window[AVGWINDOW];
	unsigned short int windowindex;
	float windowsum;
	fann_type * tmpinput;
	fann_type * tmpoutput;
	fann_type * cnvinput;
	fann_type temporary_storage;
	cnvinput=NULL;
	tmpinput=NULL;
#ifndef FLOATFANN
/* assertion that we use float as fann_type */
#error "to avoid type conflicts, the float-variant of libfann must be used"
#endif
	if (3!=argc) {
		fprintf(stderr,"EEG Artifacts Rejector -- EAR\n");
		fprintf(stderr,"version %s\n", ear_version);
		fprintf(stderr,"(c) by Andreas Romeyke, distributed under terms of GPL\n");
		fprintf(stderr,"  reads decompressed EEG-stream from STDIN, and loads a\n");
		fprintf(stderr,"  neural network. On STDOUT a REJ-stream will be written\n");
		fprintf(stderr,"call\n %s [options] <network-file> <eeg-file>\n", argv[0]);
		exit(ERRARG);
	}
	net.input=NULL;
	net.output=NULL;
	net.convert=stat;
	/* create or load neural network */
	ann = fann_create_from_file(argv[1]);
	if (ann == NULL) {
		fprintf(stderr, "could not create network from file %s: %s\n", argv[1], strerror(errno));
		exit(ERRFANN);
	}
	net.num_input=fann_get_num_input(ann); /* -1 because on-neuron for bias should not be counted */
	/* load and parse EEG-data from input */
	parse_eeg(argv[2], &net.channelcount, &net.samplecount, &net.input);
	was_an_artifact=0;
	artifact_count=0;
	for (l=0; l<AVGWINDOW; l++) {
		window[l]=0.0f;
	}
	windowindex=0;
	windowsum=0.0f;
	for (sample=0; sample< (net.samplecount-MAXTAPS); sample++) {
		tmpinput=NULL;
		tmpinput=malloc(MAXTAPS*net.channelcount*sizeof(fann_type));
		if (NULL==tmpinput) {
			fprintf(stderr, "Could not allocate memory for buffer, %s", strerror(errno));
			exit(ERRMEM);
		}
		for (taps=0; taps<MAXTAPS; taps++) {
			assert(taps<MAXTAPS);
			for (l=0; l<net.channelcount; l++) {
				assert(l<net.channelcount);
				temporary_storage=(fann_type) net.input[(sample+taps)*net.channelcount+l];
				tmpinput[taps*net.channelcount+l]=temporary_storage;
			}
		}
		/* tmp=convert_dummy((unsigned long int) (MAXTAPS * net.channelcount), &tmpinput, &cnvinput); */
		tmp=convert_stat((unsigned long int) (MAXTAPS * net.channelcount), &tmpinput, &cnvinput);
/*		tmp=convert((unsigned long int) (MAXTAPS * net.channelcount), &tmpinput, &cnvinput); */
		if (tmp != net.num_input) {
			fprintf(stderr, "count of input neurons (%u) is not equal of windowed input data (%lu), %u, %u\n",net.num_input, tmp, MAXTAPS, net.channelcount);
			exit(ERRCONVERTDATA);
		}
/*		fprintf(stderr, "tmpinput: %0x (%0x) cnvinput: %0x (%0x)\n", tmpinput, &tmpinput, cnvinput, &cnvinput);*/
		free(tmpinput);
		tmpinput=NULL;
		tmpoutput=fann_run(ann, cnvinput);
/* 		print_learncurve(sample, tmpoutput[0]); */
#ifndef NOTDEFINED
		windowsum=windowsum-window[(windowindex+1)%AVGWINDOW]+tmpoutput[0];
		window[windowindex%AVGWINDOW]= tmpoutput[0];
/*		printf("Sample:%u  out:%05f avg:%05f widx:%u %u\n",(sample+128),  tmpoutput[0], windowsum/(float)(AVGWINDOW-1), windowindex, windowindex%AVGWINDOW);*/
#ifdef DEBUG
		print_learncurve(sample, windowsum/AVGWINDOW);
#endif
		if (
			(windowsum >= AVGTHRESHOLD*(AVGWINDOW-1)) &&
			(was_an_artifact==0)
			){
			printf("%03f-", (float) (sample+OFFSET)/SAMPLEFREQUENCY);
			was_an_artifact=1;
		}
		else if (
			(windowsum < AVGTHRESHOLD*(AVGWINDOW-1))&&
			(was_an_artifact==1)
			){
			printf("%03f\n", (float) (sample+OFFSET)/SAMPLEFREQUENCY);
			was_an_artifact=0;
		}
		windowindex++;
		if (windowindex >= AVGWINDOW) windowindex=0;
#endif
#ifdef NOTDEFINED
		if (tmpoutput[0]>=0.50f)
			artifact_count++;
		else nonartifact_count++;
		if ((artifact_count > 10)&& (was_an_artifact==0)){
			printf("%03f-", (float) (sample+OFFSET)/SAMPLEFREQUENCY);
			nonartifact_count=0;
			was_an_artifact=1;
		}
		else if ((nonartifact_count > 10)&& (was_an_artifact==1)){
			printf("%03f\n", (float) (sample+OFFSET)/SAMPLEFREQUENCY);
			artifact_count=0;
			was_an_artifact=0;
		}
#endif
#ifdef NOTDEFINED
		if ((sample%1)==0 ){
			printf("sample: %u, sec: %03f EAR/EATR Test (", (sample+OFFSET), (float) (sample+OFFSET)/SAMPLEFREQUENCY);
	/*		for (i=0; i<net.channelcount; i++) {
				if (0==i) printf("%f", net.input[sample*net.channelcount+i]);
				else printf(",%f", net.input[sample*net.channelcount+i]);
			}
			*/
			/* DEBUG: to secure that all input is correct calculated */
			printf("##>");
	/*		for (taps=0; taps < MAXTAPS; taps++) {
				printf("%f ", tmpinput[taps]);
			} */
			printf("%f ", cnvinput[0]);
			/* END DEBUG: ... */
			printf(") --> ");
			/*
			for (taps=0; taps < MAXTAPS; taps++) {
				printf("%0.14f ", tmpoutput[taps]);
			}
			*/
			printf("%0.20f ", tmpoutput[0]);
			printf("\n");
		}
#endif
		(void) free_convert(&cnvinput);
	}
	if ((was_an_artifact==1) && (tmpoutput[0]<0.5f)) {
		printf("%03f\n", (float) (sample+OFFSET)/SAMPLEFREQUENCY);
		was_an_artifact=0;
	}

	(void) fflush(stdout);
	(void) fflush(stderr);
	/* destroy network */
	fann_destroy(ann);
	return(0);
}

