/** \file
	eatr - EEG artifacts train
	(c) 2004 Andreas Romeyke

	training a neural network with EEG-data

	it uses
	Fast artificial neural network library (http://fann.sourceforge.net)...

	load existing network, add new batch, outputs the learn-error, saves the network

	$Id: eatr.c,v 1.26 2004/08/05 10:02:20 romeyke Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/time.h>
#include "local_ann.h"


#define STEPSPEREPOCH 1000
/* 10000 */
static char eatr_version[]="$Id: eatr.c,v 1.26 2004/08/05 10:02:20 romeyke Exp $";
/* Prototypes */
static int help(char *argv[]);
static int calcoptions(int argc, char * argv[]);
static int dump(struct fann * ann);
static struct fann * create_network();
static void save_network(struct fann * ann);
static void destroy_network(struct fann * ann);
static void emergency_call(struct fann * ann);
int main(int argc, char * argv[]);


/* declarations */

/** calculates and checks the given options and sets the struct net */
static int calcoptions(int argc, char* argv[]) {
	int opt_case;
	unsigned int i;
	optind=1;
	opt_case=0;
	/* defaults for network parameters */
	/*	net.connection_rate=1.0f;*/
	net.connection_rate=1.0f;
	net.learning_rate=0.0001; /* 0.7f;*/
	net.num_layers=3;
	net.num_input=(MAX_MAPPED_CHANNELS*10); /* 10 is count of statistical parameters, was original (MAX_MAPPED_CHANNELS*MAXTAPS);*/
	net.num_neurons_hidden=5; /* (unsigned int) sqrt((double)net.num_input);
	if (net.num_neurons_hidden<2) net.num_neurons_hidden=2; */
	net.num_output=(1);
	net.desired_error=0.0001f;
	net.max_epochs=100000;
	net.steps_between_reports=10000;
	net.net_infile=NULL;
	net.outfilenames=NULL;
	net.net_outfile=NULL;
	net.rej_file=NULL;
	net.samplecount=0;
	net.channelcount=0;
	net.input=NULL;
	net.output=NULL;
	net.verbose=0;
	net.dump=0;
	net.net_writeperiod=net.steps_between_reports;
	net.convert=stat;
	net.adapt_learnrate=0;
	net.infiles=NULL;
	/*if (argc == 0 ) help(argv); exit(0);*/
	while (TRUE) { /* true */		opt_case=getopt(argc,argv,"ac:de:hi:l:m:o:r:v");
		if (opt_case == -1 ) {break;}
		switch (opt_case) {
			case 'a': net.adapt_learnrate=1; break;
			case 'c': net.convert=atoi(optarg); break;
			case 'd': net.dump=1; break;
			case 'e': net.desired_error=(float) atof(optarg);
						if(net.desired_error<=0) {
							fprintf(stderr, "learning error should be greater than zero\n");
							(void) help(argv);
							exit(ERRARG);
						}
						break;
			/* case 'f': net.with_fft=1; */
			case 'i': net.net_infile=optarg; break;
			case 'l': net.learning_rate=(float) atof(optarg);
						if(net.learning_rate<=0) {
							fprintf(stderr, "learning rate should be greater than zero\n");
							(void) help(argv);
							exit(ERRARG);
						}
						break;
			case 'm': net.max_epochs=(unsigned int) atoi(optarg);
						if(net.max_epochs<1) {
							fprintf(stderr, "max_epochs should be greater than zero\n");
							(void) help(argv);
							exit(ERRARG);
						}
						break;
			case 'o': net.outfilenames=optarg;
				net.net_outfile=malloc((strlen(net.outfilenames)+5)*sizeof(char));
				if (NULL==net.net_outfile) {
					fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
					exit(ERRMEM);
				}
				strcpy(net.net_outfile, net.outfilenames);
				strcat(net.net_outfile, ".net");
				net.net_mqlefile=malloc((strlen(net.outfilenames)+6)*sizeof(char));
				if (NULL==net.net_mqlefile) {
					fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
					exit(ERRMEM);
				}
				strcpy(net.net_mqlefile, net.outfilenames);
				strcat(net.net_mqlefile, ".mqle");
				net.net_mqgefile=malloc((strlen(net.outfilenames)+6)*sizeof(char));
				if (NULL==net.net_mqgefile) {
					fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
					exit(ERRMEM);
				}
				strcpy(net.net_mqgefile, net.outfilenames);
				strcat(net.net_mqgefile, ".mqge");
				net.net_cfgfile=malloc((strlen(net.outfilenames)+5)*sizeof(char));
				if (NULL==net.net_cfgfile) {
					fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
					exit(ERRMEM);
				}
				strcpy(net.net_cfgfile, net.outfilenames);
				strcat(net.net_cfgfile, ".cfg");
				break;
			case 'r': net.steps_between_reports=(unsigned int) atoi(optarg);
						if(net.steps_between_reports<1) {
							fprintf(stderr, "steps_between_reports should be greater than zero\n");
							(void) help(argv);
							exit(ERRARG);
						}
						break;
			case 'v': net.verbose=1; break;
			case '?': fprintf(stderr, "there was an unknown parameter: %i \n", optopt); (void) help(argv); exit(0);
			case ':': fprintf(stderr, "there was a required parameter missed: %s \n", optarg); (void) help(argv); exit(0);
			case 'h': (void) help(argv); exit(0);
			default: /*printf("arg was a option: %s\n",optarg);*/ break;
		}
	}
	if ((argc-optind)<1) {
		/* <1 means too few args */
		fprintf(stderr, "too few arguments, <infilenames> missed\n");
		(void) help(argv);
		exit(0);
	} else { /* parse the infiles and alloc memory for structures */
		if (NULL==net.infiles) {
			net.infiles=malloc(sizeof(struct s_eegfile)*(argc-optind));
			if (NULL==net.infiles) {
				fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
				exit(ERRMEM);
			}
		}
		assert(argc>=optind);
		net.countofinfiles=(unsigned int) argc-optind;
		for (i=0; i<net.countofinfiles; i++) {
		net.infiles[i].eeg_filename=malloc((strlen(argv[optind+i])+5)*sizeof(char));
			if (NULL==net.infiles[i].eeg_filename) {
				fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
				exit(ERRMEM);
			}
			strcpy(net.infiles[i].eeg_filename, argv[optind+i]);
			strcat(net.infiles[i].eeg_filename, ".eeg");
			net.infiles[i].rej_filename=malloc((strlen(argv[optind+i])+5)*sizeof(char));
			if (NULL==net.infiles[i].rej_filename) {
				fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
				exit(ERRMEM);
			}
			strcpy(net.infiles[i].rej_filename, argv[optind+i]);
			strcat(net.infiles[i].rej_filename, ".rej");
			net.infiles[i].chn_filename=malloc((strlen(argv[optind+i])+5)*sizeof(char));
			if (NULL==net.infiles[i].chn_filename) {
				fprintf(stderr, "could not alloc mem for buf, %s\n", strerror(errno));
				exit(ERRMEM);
			}
			strcpy(net.infiles[i].chn_filename, argv[optind+i]);
			strcat(net.infiles[i].chn_filename, ".chn");
		}
	}
	switch (net.convert) {
		case dummy: net.num_input=MAX_MAPPED_CHANNELS*MAXTAPS; break;
		case downsample:  net.num_input=MAX_MAPPED_CHANNELS*MAXTAPS/DOWNSAMPLE; break;
		case summarize_electrodes:net.num_input=MAXTAPS; break;
		case sfft:net.num_input=MAX_MAPPED_CHANNELS*MAXTAPS; break;
		case stat:net.num_input=MAX_MAPPED_CHANNELS*STATFEATURES; break;
		default: fprintf(stderr, "wrong parameter for \"-c %i\" (wrong enum_value) or error in code line: %i, file: %s\n", net.convert, __LINE__, __FILE__);
		exit(ERRARG);
	}
	return (0);
}


/** dumps the actual configuration of the ANN to a file in human-readable form */
static int dump(struct fann * ann) {
	FILE * f_dump;
	unsigned long int sumsample;
	unsigned int i;
	if (net.outfilenames!=NULL) {
		f_dump=fopen(net.net_cfgfile, "w");
		if (f_dump==NULL) {
			fprintf(stderr, "File %s could not be open, %s (line %i)\n", net.net_cfgfile, strerror(errno), __LINE__);
			exit(ERRSTREAM);
		}
	} else f_dump=stdout;
	fprintf(f_dump, "EATR version: %s\n", eatr_version);
	fprintf(f_dump, "Connection Rate: %f\n",	net.connection_rate);
	fprintf(f_dump, "Learning Rate: %f\n", net.learning_rate);
	fprintf(f_dump, "Count of Layers: %u\n", net.num_layers);
	fprintf(f_dump, "Count of Input Neurons: %u\n", net.num_input);
	fprintf(f_dump, "Count of Neurons in 1. hidden layer: %u\n", net.num_neurons_hidden);
	fprintf(f_dump, "Count of Output Neurons: %u\n", net.num_output);
	fprintf(f_dump, "Desired mean squared Learning Error: %f\n", net.desired_error);
	fprintf(f_dump, "Max Epochs: %u\n", net.max_epochs);
	fprintf(f_dump, "Samples per Epoch: %u\n", STEPSPEREPOCH );
	fprintf(f_dump, "Steps between Reports: %u\n", net.steps_between_reports);
	fprintf(f_dump, "Infile Network: %s\n", net.net_infile);
	fprintf(f_dump, "Infile Names: %u", net.countofinfiles);
	sumsample=0;
	for (i=0; i<net.countofinfiles; i++) {
		fprintf(f_dump, " (%s %s %s)", net.infiles[i].eeg_filename, net.infiles[i].rej_filename, net.infiles[i].chn_filename);
		sumsample+=net.infiles[i].samplecount;
	}
	fprintf(f_dump, "\n");
	fprintf(f_dump, "Samplecount (over all files): %lu\n", sumsample);
	fprintf(f_dump, "Outfile Names: %s\n",net.outfilenames);
	fprintf(f_dump, "Outfile Network: %s\n", net.net_outfile);
	fprintf(f_dump, "REJ-File: %s\n",net.rej_file);
	fprintf(f_dump, "adapt learnrate: %u\n", net.adapt_learnrate);
	fprintf(f_dump, "Total Neurons: %u\n", fann_get_total_neurons(ann));
	fprintf(f_dump, "Total Connections: %u\n", fann_get_total_connections(ann));
	fprintf(f_dump, "Max Taps: %u\n", MAXTAPS);
	fprintf(f_dump, "Required Channels: %u\n", MAX_MAPPED_CHANNELS);
	fprintf(f_dump, "Norm value: %u\n", NORMVALUE);
	fprintf(f_dump, "Sample Frequency: %u Hz\n",SAMPLEFREQUENCY);
	fprintf(f_dump, "Period (in Learning Steps) to save the network to disk: %u\n", net.net_writeperiod);
	fprintf(f_dump, "Convertfunction is: %i, ", net.convert);
		switch (net.convert) {
		case dummy: fprintf(f_dump,"\"dummy\"\n"); break;
		case downsample: fprintf(f_dump,"\"downsample\"\n"); break;
		case summarize_electrodes:fprintf(f_dump,"\"summarize_electrodes\"\n"); break;
		case sfft:fprintf(f_dump,"\"sfft\"\n"); break;
		case stat:fprintf(f_dump,"\"stat\"\n"); break;
		default: fprintf(stderr, "wrong parameter for \"-c %i\" (wrong enum_value) or error in code line: %i, file: %s\n", net.convert, __LINE__, __FILE__);
		exit(ERRARG);
	}
	if (net.convert==downsample) {
	fprintf(f_dump, "Downsamplingrate: %u\n", DOWNSAMPLE);
	}
	(void) fflush(f_dump);
	if (net.outfilenames!=NULL) {
		fclose(f_dump);
	}
	return(0);
}

/** prints the help-page if there were errors on the command-line or if explicitely ordered */
static int help(char * argv[]) {
	fprintf(stderr,"EEG Artifacts Training -- EATR\n");
	fprintf(stderr,"(c) by Andreas Romeyke, distributed under terms of GPL\n");
	fprintf(stderr,"version: %s\n", eatr_version);
	fprintf(stderr,"  reads decompressed EEG-stream from STDIN, and trains a\n");
	fprintf(stderr,"  neural network with the correspondent rej-data. The network will be saved,\n");
	fprintf(stderr,"  if max_epochs will be reached or actual\n");
	fprintf(stderr,"  error below desired_error. The network will be saved.\n");
	fprintf(stderr,"call\n %s [options] <filename-group>\n", argv[0]);
	fprintf(stderr,"or\n %s -h\n",argv[0]);
	fprintf(stderr,"options are:\n");
	fprintf(stderr, "\t-a\tadaptr learnrate, means per epoch the learningrate decrease (default:%u)\n", net.adapt_learnrate);
	fprintf(stderr, "\t-c VAL\tconvert function (int, default: %i)\n", net.convert);
	fprintf(stderr, "\t-e VAL\tlearning error over an epoch (float, default: %03f)\n", net.desired_error);
	fprintf(stderr, "\t-h    \thelp\n");
	fprintf(stderr, "\t-i VAL\tnet file to load (string, default: %s)\n", net.net_infile);
	fprintf(stderr, "\t-l VAL\tlearning_rate (float, default:%0.10f)\n", net.learning_rate);
	fprintf(stderr, "\t-m VAL\tmax_epochs (unsigned int, default: %u)\n", net.max_epochs);
	fprintf(stderr, "\t-o VAL\toutfiles (outfile names without extensions) to store (string, default:%s)\n",net.outfilenames);
	fprintf(stderr, "\t-r VAL\tsteps_between_reports (unsigned int, default: %u)\n", net.steps_between_reports);
	fprintf(stderr, "\t-v\t verbose (default: %u)\n", net.verbose);
	fprintf(stderr, "\n");
	return(0);
}

/** small function to encapsulate the creation/loading of ANN */
static struct fann * create_network() {
	struct fann * ann;
	/* create or load neural network */
	if (NULL != net.net_infile) { /* load network-data from file */
		ann = fann_create_from_file(net.net_infile);
		if (ann == NULL) {
			fprintf(stderr, "could not create network from file %s: %s\n", net.net_infile, strerror(errno));
			exit(ERRFANN);
		}
	} else { /* build network from data in net-structure */
		ann = fann_create(net.connection_rate, net.learning_rate, net.num_layers,net.num_input, net.num_neurons_hidden, net.num_output);
		if (ann == NULL) {
			fprintf(stderr, "could not create network: %s\n", strerror(errno));
			exit(ERRFANN);
		}
		fann_randomize_weights(ann, -5.0f, 5.0f);
		/* fann_set_momentum(ann, 0.9); */
		fann_set_training_algorithm(ann, FANN_TRAIN_QUICKPROP);

		/* fann_set_activation_function_output(ann, FANN_THRESHOLD); */
		/* fann_set_activation_function_hidden(ann, FANN_THRESHOLD); */
		fann_set_activation_function_output(ann, FANN_SIGMOID);
		fann_set_activation_function_hidden(ann, FANN_SIGMOID);
		/*fann_set_activation_function_output(ann, FANN_SIGMOID_SYMMETRIC);
		fann_set_activation_function_hidden(ann, FANN_SIGMOID_SYMMETRIC); */
	}
	return(ann);
}

/** small function to encapsulate the peridocal backup mechanism to store the actual trained ANN */
static void save_network(struct fann * ann) {
	assert(ann !=NULL);
	/* save network */
	if (net.net_outfile!=NULL) { /* only if filename specified... */
		fann_save(ann,net.net_outfile);
	}
}

/** small function to free the ANN structure */
static void destroy_network(struct fann * ann) {
	assert(ann!=NULL);
	fann_destroy(ann);
}

/** to free stuff on programs exit */
static void emergency_call(struct fann * ann) {
	/* save network */
	save_network(ann);
	/* free mem */
	destroy_network(ann);
	if (NULL!=net.output)
		free_rej(&net.output);
	if (NULL!=net.input)
		free_eeg(&net.input);
}

/** main logic to prepare ANN and feed it with the right data to train an ANN */
int main(int argc, char * argv[]) {
	FILE * f_mqle;
	FILE * f_mqge;
	struct fann * ann;
	unsigned int i,j,k,l,selected_infile;
	struct timeval lasttimestamp;
	struct timeval timestamp;
	unsigned short int taps;
	unsigned long int tmp;
	unsigned int artifactness;
	double artifact_ratio;
	fann_type * tmpinput;
	fann_type * cnvinput;
	fann_type * tmpoutput;
	fann_type * testoutput;
	fann_type temporary_storage;
	fann_type epoch_learning_error;
	fann_type generalization_error;
	unsigned long generalization_counter;
	fann_type tmperror;
	cnvinput=NULL;
	artifact_ratio=0.0;
	generalization_counter=0;
	gettimeofday(&timestamp,NULL);
	lasttimestamp.tv_sec=timestamp.tv_sec;
	lasttimestamp.tv_usec=timestamp.tv_usec;
 #ifndef FLOATFANN
 /* assertion that we use float as fann_type */
 #error "to avoid type conflicts, the float-variant of libfann must be used"
 #endif
	(void) calcoptions(argc, argv);
	if (net.outfilenames!=NULL) {
		f_mqle=fopen(net.net_mqlefile, "w");
		if (f_mqle==NULL) {
			fprintf(stderr, "File %s could not be open, %s (line %i)\n", net.net_mqlefile, strerror(errno), __LINE__);
			exit(ERRSTREAM);
		}
		fprintf(f_mqle, "# sample, f_mqle over %u steps, time in s\n", net.steps_between_reports);
		f_mqge=fopen(net.net_mqgefile, "w");
		if (f_mqge==NULL) {
			fprintf(stderr, "File %s could not be open, %s (line %i)\n", net.net_mqgefile, strerror(errno), __LINE__);
			exit(ERRSTREAM);
		}
		fprintf(f_mqge, "# sample, f_mqge over %i steps\n", STEPSPEREPOCH);
	} else {
		f_mqge=stdout;
		f_mqle=stderr;
	}

	ann=create_network();
	srand((unsigned int) time(NULL));
	for (i=0; i<net.countofinfiles; i++) {
		/* load all eeg-data into buffers */
		(void) parse_eeg(net.infiles[i].eeg_filename,&net.infiles[i].channelcount, &net.infiles[i].samplecount, &net.infiles[i].eeg_buffer);
		/* parse and load all rejection-data into buffers */
		(void) parse_rej(net.infiles[i].rej_filename, &net.infiles[i].samplecount, &net.infiles[i].rej_buffer);
		if (0==i) net.channelcount=net.infiles[i].channelcount;
		if (net.channelcount != net.infiles[i].channelcount) {
			fprintf(stderr, "the input files have different channelcounts\n");
			exit(ERRDIFFFILES);
		}
	}
	if (1==net.dump) { /* dump to STDOUT for a quickdump */
		(void) dump(ann);
		emergency_call(ann);
		exit(0);
	} else if (NULL==net.outfilenames) { ; /* do nothing */
	} else (void) dump(ann); /* dump to file */
	/* train network */
	fprintf(stderr, "total neurons: %u, total connections: %u\n", fann_get_total_neurons(ann), fann_get_total_connections(ann));
	fprintf(stderr, "start with training\n"); (void) fflush(stderr);
	artifactness=0;
	/* from here the learning will be described */
	for (i=0; i< net.max_epochs; i++) {
		epoch_learning_error=1.0f;
		if (1==net.adapt_learnrate && i>0) { /* adapt the learnrate */
			if (fann_get_learning_rate(ann) > 0.00000001) { /* a value less than this makes no sense */
				fann_set_learning_rate(ann,
					0.99*fann_get_learning_rate(ann)
				);
			}
		} else {
			fann_set_learning_rate(ann,net.learning_rate);
		}
		generalization_error=0.0f;
		generalization_counter=0;

		for (j=1; j<STEPSPEREPOCH; j++) { /* learning-steps */
			/* next, choose one inout/output randomly */
			fann_reset_MSE(ann); /* reset mean learning error */
			selected_infile=mrandomize(net.countofinfiles);
			assert(net.infiles[selected_infile].samplecount>MAXTAPS);
			/* create a new temporary field to store input data with taps */
			tmp=0;
			tmpoutput=NULL;
			tmpoutput=malloc(1*sizeof(fann_type));
			if (NULL==tmpoutput) {
				fprintf(stderr, "could not allocate memory for buffer\n, %s", strerror(errno));
				exit(ERRMEM);
			}
			tmpinput=NULL;
			tmpinput=malloc(MAXTAPS*MAX_MAPPED_CHANNELS*sizeof(fann_type));
			if (NULL==tmpinput) {
				fprintf(stderr, "could not allocate memory for buffer\n, %s", strerror(errno));
				exit(ERRMEM);
			}
			tmpoutput[0]=0.0f;
			k=mrandomize(net.infiles[selected_infile].samplecount-MAXTAPS); /* to have a time-series */
			assert(k<=(net.infiles[selected_infile].samplecount-MAXTAPS));
			/* calculate the count of artifact-samples in actual window */
			for (taps=0; taps<MAXTAPS; taps++) {
				assert(net.countofinfiles+1 > 0);
				tmpoutput[0]+=net.infiles[selected_infile].rej_buffer[k+taps];
			}
			assert(MAXTAPS>0);
			/* use the count to calculate the "artifactness" of actual window */
			tmpoutput[0]/=(float)MAXTAPS;
			for (taps=0; taps<MAXTAPS; taps++) {
				assert(taps<MAXTAPS);
				tmpinput[taps*net.channelcount]=0.0f;			/* DONE: tmpinput must be init with 0 */
				for (l=0; l<net.channelcount; l++) { /* create input for net */
					assert(l<net.channelcount);
					/* DEBUG: print calculation of adress
					fprintf(stderr, "taps: %i, maxchann: %i, l(chan): %i, k(rand): %i, maxk: %i -- left: %x, right %x\n", taps, net.channelcount, l, k, net.samplecount-MAXTAPS, taps*net.channelcount+l, (k+taps)*net.channelcount+l);
					*/
					temporary_storage=(fann_type) net.infiles[selected_infile].eeg_buffer[(k+taps)*net.infiles[selected_infile].channelcount+l];
					tmpinput[taps*net.channelcount+l]=temporary_storage;
				}
			}
			/**** tmp=convert_stat((unsigned long int) (MAXTAPS * net.infiles[selected_infile].channelcount), &tmpinput, &cnvinput); */
			tmp=convert((unsigned long int) (MAXTAPS * net.infiles[selected_infile].channelcount), &tmpinput, &cnvinput);
/*			tmp=convert_downsample((unsigned long int) (MAXTAPS * net.channelcount), &tmpinput, &cnvinput); */
			if (tmp != net.num_input) {
				fprintf(stderr, "count of input neurons (%u) is not equal of windowed input data (%lu)\n",net.num_input, tmp);
				exit(ERRCONVERTDATA);
			}
			artifact_ratio+=tmpoutput[0];
			if (((tmpoutput[0]>0.6f)&&(artifactness==0))|| /* the value should be the half of a typical artifact. If the typical artifact-length==0,5 sec and Samplefrequency = 250 Hz and the MAXTAPS=250, then the value should be 0.25 */
				((tmpoutput[0]<=0.4f)&&(artifactness==1))){
				artifactness=1 ^ artifactness;
				testoutput=fann_run(ann, cnvinput);
				tmperror=0.0f;
				tmperror+= ((tmpoutput[0]-testoutput[0])*(tmpoutput[0]- testoutput[0])); /* generalization */
				generalization_error+=(tmperror);
				generalization_counter++;
				fann_train(ann, cnvinput, tmpoutput);
				/* TODO: fann_save_train(); */
			} else {
			}
			(void) free_convert(&cnvinput);
			free(tmpinput);
			free(tmpoutput);
			tmpinput=NULL;
			tmpoutput=NULL;
			if ((0==((i*STEPSPEREPOCH+j-1)%net.steps_between_reports))&&
				(net.outfilenames!=NULL)) {
				gettimeofday(&timestamp,NULL);

				fprintf(f_mqle, "%u %0.10f %0.10f\n", (i*STEPSPEREPOCH+j-1),fann_get_MSE(ann), fann_get_learning_rate(ann));

				 /* to trace learning with gnuplot, per epoch */
				 (void) fflush(f_mqle);
			}
			if ((net.net_outfile!=NULL) &&
				(0 ==  ((i*STEPSPEREPOCH+j-1)%net.net_writeperiod))) { /* only if filename specified... */
				save_network(ann);
			}
			(void) fflush(stderr);
		}
		if (net.outfilenames!=NULL) {
			 fprintf(f_mqge, "%u %0.10f\n", (i*STEPSPEREPOCH+j-1),(generalization_error/((double)1+generalization_counter)));
			 fflush(f_mqge);
		}
		/*if ((fann_get_MSE(ann) < net.desired_error) || */
	}
	artifact_ratio/= (double) (i*STEPSPEREPOCH+j);
	fprintf(stderr, "Artifact-ratio was: %f\n", artifact_ratio);
	save_network(ann);
	destroy_network(ann);
	if (net.outfilenames!=NULL) {
		fclose(f_mqle);
		fclose(f_mqge);
	}
	return(0);
}

