#ifndef _MY_ANN_LOCAL_H
#define _MY_ANN_LOCAL_H


/** libfann */
#include "floatfann.h"

/** project wide definitions */
#include "local.h"


/** holds the global configuration of ANN and related stuff in eatr.c and co. */
struct s_net {
	/** 1 means full connected, 0.5 half connected... */
	float connection_rate;
	/** 0.7 is a good choice */
	float learning_rate;
	/** 3 means, one input, one hidden, one output, 3 is a good choice */
	unsigned int num_layers;
	/** equivalent to (Channels * TAPs) in EEG-data, should be 64*20=1280 */
	unsigned int num_input;
	/** should be equivalent to count of channels */
	unsigned int num_neurons_hidden;
	/**  should be 1, because it signals that was an artefact or not */
	unsigned int num_output;
	/** abort the training if learning error is less desired error  */
	float desired_error;
	/** abort if max_epochs reached, it doesnot mean epoch-learning, its pattern-learning! */
	unsigned int max_epochs;
	/** every "epochs_between_report" epoche generate a report about state and statistics */
	unsigned int steps_between_reports;
	/** rej-file */
	char * rej_file;
	/** net-file to load network from */
	char * net_infile;
	/** general prefix for files to be written */
	char * outfilenames;
	/** net-file to store network to */
	char * net_outfile;
	/** net-file to store mean square learning error to */
	char * net_mqlefile;
	/** net-file to store mean square generalization error to */
	char * net_mqgefile;
	/** net-file to store network config to (machine parseable) */
	char * net_cfgfile;
	/** count of infiles */
	unsigned int countofinfiles;
	/** array of general prefixes for files to be loaded */
	struct s_eegfile * infiles;
	unsigned int samplecount;
	unsigned short int channelcount;
	float * input;
	unsigned short * output;
	unsigned short int verbose;
	unsigned short int dump;
	/* unsigned short int with_fft; */
	unsigned short int adapt_learnrate;
	/** after which steps the network should be periodically saved */
	unsigned int net_writeperiod;
	enum e_convert convert;
} net; /* globally holds parameters of network */


/* shared/convert.c */
void free_convert(fann_type ** ppout);
unsigned long int convert_dummy(unsigned long int size_in, fann_type ** ppin, fann_type ** ppout);
unsigned long int convert_shorttime_fft(unsigned long int size_in, fann_type ** ppin, fann_type ** ppout);
unsigned long int convert_summarize_electrodes(unsigned long int size_in, fann_type ** ppin, fann_type ** ppout);
unsigned long int convert_downsample(unsigned long int size_in, fann_type ** ppin, fann_type ** ppout);
unsigned long int convert_stat(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout);
unsigned long int convert(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout);
#endif
