/** \file
	parse_eeg
	2004, Andreas Romeyke
	mailto://art1@it-netservice.de

*/

#include "local.h"

/** parse_eeg parses the decompressed data from programm "cntconvert". */
int parse_eeg(char * filename, unsigned short int * chanc, unsigned int * samplec, float ** ppebuf){
	FILE * EIN;
	float * pebuf;
	struct s_cntconvhead header;
	int ret;
	char string[32];
	/* FILE * fstdin; */
	assert(ppebuf != NULL); /* should be real mem */
	assert(samplec != NULL); /* should be real mem */
	pebuf= *ppebuf;
	assert(pebuf == NULL); /* should not point to memory */
	assert(chanc!=NULL); /* should a real address */
	assert(samplec!=NULL); /* should a real address */
	/* read from file */
	EIN=fopen(filename, "rb");
	if (EIN == NULL) {
		fprintf(stderr, "could not open file \"%s\"\n", filename);
		exit(ERRSTREAM);
	}
	ret=fscanf(EIN, "%6s", string);
	if (1 != ret) {
		fprintf(stderr, "wrong input data from file %s, %i, %s, %s, %u\n", filename, ret, strerror(errno), __FILE__, __LINE__);
		exit(ERRSTREAM);
	}
	if (0!=strcmp(string, "#EEGbu")) {
		fprintf(stderr, "wrong input data from file %s, %s, %u\n", filename, __FILE__, __LINE__);
#ifndef NDEBUG
		fprintf(stderr, "string \"%s\" differs from filestring \"%s\" at pos %u\n","#EEGbu", string, (unsigned int) strspn("#EEGbu", string));
#endif
		exit(ERRSTREAM);
	}
	/* TODO: for higher portability we should not read/write the structure at once, but its elements, check also cntconv.c */
	ret=(int) fread(&header, sizeof(header), 1,EIN);
	if (1 != ret) {
		fprintf(stderr, "wrong input data from file \"%s\" %i, %s, %u\n", filename, ret, strerror(errno), __LINE__);
		exit(ERRSTREAM);
	}
	assert(header.channelcount>0);
	assert(header.samplecount>0);

	pebuf=(float *) malloc((header.samplecount * header.channelcount)*sizeof(float));
	if (NULL == pebuf) {
		fprintf(stderr, "Could not allocate memory for buffer pebuf, %u,%u, %u MB\n", header.samplecount, header.channelcount, (unsigned int) (header.channelcount * header.samplecount * sizeof(float))/(1024*1024));
		exit(ERRMEM);
	}
	ret=(int)fread(pebuf,sizeof(float), (size_t) (header.samplecount * header.channelcount), EIN);
	if ((int) (header.samplecount * header.channelcount) != ret) {
		fprintf(stderr, "wrong input data from file \"%s\" %i, %s, (%u, %u) %s, %u\n", filename,  ret, strerror(errno), (unsigned int)sizeof(float), header.samplecount * header.channelcount, __FILE__, __LINE__);
		exit(ERRSTREAM);
	}
	*ppebuf=pebuf;
	*chanc=(unsigned short int) header.channelcount;
	*samplec=header.samplecount;
	return(0);
}

/** free the by parse_eeg() allocated memory */
void free_eeg(float ** ppebuf) {
	float * pebuf;
	assert(ppebuf != NULL); /* should be real mem */
	pebuf= *ppebuf;
	assert(pebuf != NULL); /* should point to memory */
	free(pebuf);
	pebuf=NULL;
	*ppebuf =pebuf;
}


#ifdef EEG_TEST
int main (int argc, char * argv[]) {
	unsigned short int i;
	unsigned int j;
	float * pebuf;
	pebuf=NULL;
	parse_eeg("testdata.eeg",&i,&j, &pebuf);
	free_eeg(&pebuf);
	return(0);
}
#endif

