/** \file
	parse_rej
	2004, Andreas Romeyke
	mailto://art1@it-netservice.de

*/

#include "local.h"

/** parse_rej parses the rej-Files. The problem is, the values are based on seconds instead samples. This function convert it to samples based on a sampling frequency of 250 Hz */
int parse_rej(char * filename, unsigned int * samplec, unsigned short int ** pprbuf){
	FILE * RIN;
	float begintime;
	float endtime;
	char dummy;
	int ret;
	unsigned int i;
	int actline;
	unsigned int sample;
	unsigned int samplecount;
	unsigned short int * prbuf;
	assert(pprbuf != NULL); /* should be real mem */
	assert(samplec != NULL); /* we need the real value of samplecount to alloc mem */
	prbuf= *pprbuf;
	assert(prbuf == NULL); /* should not point to memory */
	samplecount = *samplec;
	assert(samplecount>0); /* we need at least one value */
	actline=0;
	/* open the rej */
	if (NULL==filename) {
		fprintf(stderr, "there was no filename specified for REJ-data, %s\n", strerror(errno));
		exit(ERRSTREAM);
	}
	RIN = fopen(filename, "r");
	if (RIN == NULL) {
		fprintf(stderr, "cannot open %s, %s\n", filename, strerror(errno));
		exit(ERRSTREAM);
	}
	prbuf=(unsigned short int *) malloc(sizeof(unsigned short int)*samplecount);
	if (prbuf == NULL) {
		fprintf(stderr, "Could not allocate memory for buffer\n");
		(void) fclose(RIN);
		exit(ERRMEM);
	}
	/* init prbuf to '0' means "sample is no artifact" */
	for (i=0; i<samplecount; i++) {
		prbuf[i]=0;
	}

	/* read rej-File and fill prbuf with artifact-data */
	sample=0;
	while (
		(feof(RIN)!= -1 )||
		(sample>samplecount)
		) {
		ret=fscanf(RIN, "%f%c%f",&begintime, &dummy, &endtime);
		actline++;
		if (ret==EOF) break;
		if (ret!=3) {
			fprintf(stderr, "file %s has wrong data at line %i\n", filename, actline);
			(void) fclose(RIN);
			exit(ERRSTREAM);
		}
		for (sample=(unsigned int)(begintime*SAMPLEFREQUENCY); sample<(unsigned int) (endtime*SAMPLEFREQUENCY); sample++) { /* ok */
			if (sample>samplecount) {
				fprintf(stderr, "file %s at line %i describes a longer timeperiod than EEG-data (with %us)\n", filename, actline, samplecount/SAMPLEFREQUENCY);
				(void) fclose(RIN);
				exit(ERRSTREAM);
			}
			prbuf[sample]=1;
		}
	}
	(void) fclose(RIN);
	*pprbuf = prbuf; /* return data */
	return(0);
}


/** free the by parse_rej() allocated memory */
void free_rej(unsigned short int ** pprbuf) {
	unsigned short int * prbuf;
	assert(pprbuf != NULL); /* should be real mem */
	prbuf= *pprbuf;
	assert(prbuf != NULL); /* should point to memory */
	free(prbuf);
	prbuf=NULL;
	*pprbuf=prbuf;
}

#ifdef EEG_TEST
/** to test the correct function of parse_rej() */
int main (int argc, char * argv[]) {
	int i, j;
	float * pebuf;
	pebuf=NULL;
	if (argc!=3) {
		fprintf(stderr, "call testprogram: %s <SAMPLECOUNT> <FILE.REJ>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	i=atoi(argv[1]);
	printf("you gave me: samplecount=%i rej-File=%s\n", i, argv[2]);
	parse_rej(argv[1],&j, &pebuf);
	free_rej(&pebuf);
	return(0);
}
#endif
